﻿using System;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.CommonBot;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Action = Styx.TreeSharp.Action;
using HkM = FmBalance.Core.Managers.HotKeysManager;
using L = FmBalance.Core.Utilities.Log;

namespace FmBalance.Core
{
    static class Spell
    {
        public delegate WoWUnit unitSelectionDelegate(object context);



        public delegate T Selection<out T>(object context);

        public static int lastSpellCast;
        private static LocalPlayer Me => StyxWoW.Me;
        private static WoWUnit currentTarget => StyxWoW.Me.CurrentTarget;

        #region [Method] - Cast Method

        private static Composite castMethod(this int Spell, unitSelectionDelegate unitSelection,
            Selection<bool> reqs = null, bool gcdCast = false)
        {
            return
                new Decorator(
                    ret =>
                        ((reqs != null && reqs(ret)) || (reqs == null)) && unitSelection != null &&
                        unitSelection(ret) != null && SpellManager.CanCast(Spell, unitSelection(ret)),
                    new Action(ret =>
                    {
                        if (SpellManager.Cast(Spell, unitSelection(ret)))
                        {
                            lastSpellCast = Spell;
                            L.combatLog("Casting: " + WoWSpell.FromId(Spell).Name + " on " + unitSelection(ret).SafeName);
                            if (!gcdCast)
                                return RunStatus.Success;
                        }
                        return RunStatus.Failure;
                    }));
        }
        private static Composite castMethod(this string Spell, unitSelectionDelegate unitSelection,
    Selection<bool> reqs = null, bool gcdCast = false)
        {
            return
                new Decorator(
                    ret =>
                        ((reqs != null && reqs(ret)) || (reqs == null)) && unitSelection != null &&
                        unitSelection(ret) != null && SpellManager.CanCast(Spell, unitSelection(ret)),
                    new Action(ret =>
                    {
                        if (SpellManager.Cast(Spell, unitSelection(ret)))
                        {
                            SpellFindResults sfr;
                            SpellManager.FindSpell(Spell, out sfr);
                            WoWSpell spellid = sfr.Override ?? sfr.Original;
                            lastSpellCast = spellid.Id;
                            L.combatLog("Casting: " + Spell + " on " + unitSelection(ret).SafeName);
                            if (!gcdCast)
                                return RunStatus.Success;
                        }
                        return RunStatus.Failure;
                    }));
        }


        #endregion

        #region Drop Method

        private static Composite dropMethod(this int Spell, unitSelectionDelegate unitSelection, Selection<bool> reqs = null, bool waitForSpell = false)
        {
            return new Decorator(ret => unitSelection != null && reqs(ret) && SpellManager.CanCast(Spell) && (Me.Location.Distance(unitSelection(ret).Location) <= WoWSpell.FromId(Spell).MaxRange || WoWSpell.FromId(Spell).MaxRange == 0),
            new Sequence(
            new Action(ret => SpellManager.Cast(Spell)),
            new DecoratorContinue(ctx => waitForSpell,
            new WaitContinue(1, ret => Me.CurrentPendingCursorSpell != null && Me.CurrentPendingCursorSpell.Id == Spell, new ActionAlwaysSucceed())),
            new Action(ret =>
            {
                SpellManager.ClickRemoteLocation(unitSelection(ret).Location);
                HkM.switchKeyValue(Spell);
                lastSpellCast = Spell;
                L.combatLog("Casting: " + Spell);
            })));
        }

        private static Composite dropMethod(this string Spell, unitSelectionDelegate unitSelection, Selection<bool> reqs = null, bool waitForSpell = false)
        {
            SpellFindResults sfr;
            SpellManager.FindSpell(Spell, out sfr);
            WoWSpell spellid = sfr.Override ?? sfr.Original;
            return new Decorator(ret => unitSelection != null && reqs(ret) && SpellManager.CanCast(Spell) && (Me.Location.Distance(unitSelection(ret).Location) <= spellid.MaxRange || spellid.MaxRange == 0),
            new Sequence(
            new Action(ret => SpellManager.Cast(Spell)),
            new DecoratorContinue(ctx => waitForSpell,
            new WaitContinue(1, ret => Me.CurrentPendingCursorSpell != null && Me.CurrentPendingCursorSpell == spellid, new ActionAlwaysSucceed())),
            new Action(ret =>
            {
                SpellManager.ClickRemoteLocation(unitSelection(ret).Location);
                lastSpellCast = spellid.Id;
                L.combatLog("Casting: " + Spell);
            })));
        }
        #endregion

        #region [Method] - Cooldown Detection

        public static double cooldownTimeLeft(this int Spell)
        {
            try
            {
                SpellFindResults Results;
                return SpellManager.FindSpell(Spell, out Results)
                    ? (Results.Override != null
                        ? Results.Override.CooldownTimeLeft.TotalMilliseconds
                        : Results.Original.CooldownTimeLeft.TotalMilliseconds)
                    : 9999;
            }
            catch (Exception xException)
            {
                L.diagnosticLog("Exception in cooldownTimeLeft(): ", xException);
                return 9999;
            }
        }

        #endregion

        #region [Method] - Ground Casting
        public static Composite dropCast(this int Spell, unitSelectionDelegate unitSelection, Selection<bool> reqs = null)
        {
            return dropMethod(Spell, unitSelection, reqs);
        }
        public static Composite dropCast(this string Spell, unitSelectionDelegate unitSelection, Selection<bool> reqs = null)
        {
            return dropMethod(Spell, unitSelection, reqs);
        }
        #endregion

        #region [Method] - Standard Casting

        public static Composite Buff(this int Spell, Selection<bool> reqs = null, bool gcdCast = false)
        {
            return castMethod(Spell, ret => Me, reqs, gcdCast);
        }

        public static Composite Cast(this int Spell, Selection<bool> reqs = null, bool gcdCast = false)
        {
            return castMethod(Spell, ret => currentTarget, reqs, gcdCast);
        }
        public static Composite Cast(this string Spell, Selection<bool> reqs = null, bool gcdCast = false)
        {
            return castMethod(Spell, ret => currentTarget, reqs, gcdCast);
        }
        public static Composite Cast(this string Spell, unitSelectionDelegate onUnit, Selection<bool> reqs = null, bool gcdCast = false)
        {
            return castMethod(Spell, onUnit, reqs, gcdCast);
        }


        #endregion

        #region Spell Charges

        public static int GetCharges(string name)
        {
            SpellFindResults sfr;
            if (SpellManager.FindSpell(name, out sfr))
            {
                WoWSpell spell = sfr.Override ?? sfr.Original;
                return GetCharges(spell);
            }
            return 0;
        }

        public static int GetCharges(WoWSpell spell)
        {
            int charges = Lua.GetReturnVal<int>("return GetSpellCharges(" + spell.Id.ToString() + ")", 0);
            return charges;
        }
        #endregion

        public static float SpellDistance(this WoWUnit unitOrigin, WoWUnit unitTarget = null)
        {
            // abort if mob null
            if (unitOrigin == null)
                return 0;

            // when called as SomeUnit.SpellDistance()
            // .. convert to SomeUnit.SpellDistance(Me)
            if (unitTarget == null)
                unitTarget = StyxWoW.Me;

            // when called as SomeUnit.SpellDistance(Me) then
            // .. convert to Me.SpellDistance(SomeUnit)
            if (unitTarget.IsMe)
            {
                unitTarget = unitOrigin;
                unitOrigin = StyxWoW.Me;
            }

            // only use CombatReach of destination target
            float dist = unitTarget.Location.Distance(unitOrigin.Location) - unitTarget.CombatReach;
            return Math.Max(0, dist);
        }

        public static bool MaxChargesAvailable(int Spell)
        {
            SpellChargeInfo sci = GetSpellChargeInfo(Spell);
            if (sci != null)
            {
                return (sci.ChargesLeft == sci.MaxCharges);
            }
            return false;
        }

        public static SpellChargeInfo GetSpellChargeInfo(int Spell)
        {
            try
            {
                SpellFindResults Results;
                SpellManager.FindSpell(Spell, out Results);
                return (Results.Override != null ? Results.Override.GetChargeInfo() : Results.Original.GetChargeInfo());

            }
            catch (System.Exception xException)
            {
                return null;
            }
        }


        public static SpellChargeInfo GetSpellChargeInfo(string Spell)
        {
            try
            {
                SpellFindResults Results;
                SpellManager.FindSpell(Spell, out Results);
                return (Results.Override != null ? Results.Override.GetChargeInfo() : Results.Original.GetChargeInfo());

            }
            catch (System.Exception xException)
            {
                return null;
            }
        }
    }

    
}
